#include <stdlib.h>
#include <stdio.h>
#include <omp.h>
#include <structs.h>
#include <math.h>
#include <fstream>
#include <string.h>
#include <linear_int.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_bessel.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <myspline.h>
#include <relid3D.h>



#include <gsl/gsl_multimin.h>
#include <gsl/gsl_rng.h>

using namespace std;

double ETAOS=0.1;

double alpha=150.;
double regulator=0.05;
double regsigma=0.01;
double bulkreg=0.0;
double a10=0.0; //was -4
double dof=1; //number of degrees of freedom

fstream meta;

int ORDER=2; //order to stepping algorithm
const int LORDER=1;

const int CDIM=3;

const int REFZ=1.0;

const double EDreg=1.e-10;
const double Ureg=1.;

int SNAPOUT=40;

double dt=0.1; //ratio of time step over lattice spacing

char cbuffer[1255];
char *outdir;

//#define INTERPOLATE 1;
//#define SIMPLE 1;
//#define UPWIND 1;
#define MINMOD 1;


//runs nicely
//#define BAYMMODIC 1
double AT=0.633463/2.;
int NX=439;
int NY=439;
int NZ=1;
double TINIT=1.0; //initialization time in fm/c


int debug = 0;
const int OUTPUTWARN = 0;
int ABORT = 0;
int NEGFEQWARN = 0;
int NEGFWARN = 0;
int NANWARN =0;

const double fmtoGeVI=5.0677;

int num_threads=1; //number of threads to use for OMP

char desc[255];

//#define D3Q1_6_3 1;
//#define D3Q1_6_3_plus 1;
//#define D3Q1_8_3 1;
//#define D3Q1_8_3_plus 1;
//#define D3Q1_8_4_plus 1;
//#define D3Q1_8_5_plus 1;
//#define D3Q18 1;
#define D3Q26 1;
//#define MILNE 1;



#ifdef D3Q1_6_3
const int SPEEDS=18;
const int numNpoly=9;
#endif 

#ifdef D3Q1_6_3_plus
const int SPEEDS=19;
const int numNpoly=10;
#define plus 1;
#endif 

#ifdef D3Q1_8_3
const int SPEEDS=24;
const int numNpoly=9;
#endif 


#ifdef D3Q1_8_3_plus
const int SPEEDS=25;
const int numNpoly=10;
#define plus 1;
#endif 

#ifdef D3Q1_8_4_plus
const int SPEEDS=33;
const int numNpoly=11;
#define plus 1;
#endif

#ifdef D3Q1_8_5_plus
const int SPEEDS=41;
const int numNpoly=11;
#define plus 1;
#endif

#ifdef D3Q18
const int SPEEDS=18;
const int numNpoly=10;
#endif

#ifdef D3Q26
const int SPEEDS=26;
const int numNpoly=10;
#endif



//#ifndef INITFILE
#define INITGAUSSIAN 1;
//#define BAYMMODIC 1;
//#define FS1D 1;
//#endif
//#define INITHANS


xygrid *f; //distribution function
xygrid *fnew; //updated distribution function
xygrid *Tab; //local Energy-momentum tensor
xygrid *T;//local temperature
xygrid *v;//local 3-velocity
xygrid *u0;//local u^0
int *warn;//warnings
double c[CDIM+1][SPEEDS]; //velocity vectors c_t,c_x,c_y,c_z
double dist[SPEEDS]; //distance to NN
double w[SPEEDS]; //velocity vectors weights
double wtw[SPEEDS][CDIM+1][CDIM+1]; //velocity vectors weights including exp(z0*c0)*c[a]*c[b];
double poly[numNpoly][SPEEDS]; //orthogonal polynomials
double Npoly[numNpoly]; //normalization of polynomials
double t=0; //global time step
myspline logeot,enthalpy;


#ifdef MILNE
double dzpoly[numNpoly][SPEEDS]; //d/dz^v derivative of orthogonal polynomials
#endif

void allocate_memory(xygrid * XCgrid)
{
  printf("===> Info: Allocating memory...\t");

  f = new xygrid[XCgrid->SIZE];
  fnew = new xygrid[XCgrid->SIZE];
  Tab = new xygrid[XCgrid->SIZE];
  T = new xygrid[XCgrid->SIZE];
  v = new xygrid[XCgrid->SIZE];
  u0 = new xygrid[XCgrid->SIZE];
  warn = new int[XCgrid->SIZE];
  for(long int i=0;i<XCgrid->SIZE;i++)
    {
      f[i].init(SPEEDS,1,1);
      f[i].alloc_data();
      Tab[i].init(4,4,1);
      Tab[i].alloc_data();
      T[i].init(1,1,1);
      T[i].alloc_data();
      v[i].init(3,1,1);
      v[i].alloc_data();
      u0[i].init(1,1,1);
      u0[i].alloc_data();   
      fnew[i].init(SPEEDS,1,1);
      fnew[i].alloc_data();
    }
  printf("\t finished\n");
}

void free_memory()
{
  printf("===> Info: Freeing  memory...\t");

  delete [] f;
  delete [] fnew;
  delete [] Tab;
  delete [] T;
  delete [] v;
  delete [] u0;
  delete [] warn;
  printf("finished\n");
}

double max(double a,double b)
{
  if (a>b)
    return a;
  else
    return b;
}

double minmod(double a,double b)
{
  double res=0.;
  if (a*b>0)
    {
      if (fabs(a)<fabs(b))
	res=a;
      else
	res=b;
    }
  return res;
}


void setspeeds()
{

  printf("===> Info: Setting %i Lattice Speeds...\t",SPEEDS);



#ifdef D3Q1_6_3

  //for z=1.0

  double vx []={0.6324555320336758, 1., 0.632455532033676, 0.3162277660168379, 0.5, 
 0.316227766016838, -0.3162277660168379, -0.5, -0.316227766016838, 
 -0.6324555320336758, -1., -0.632455532033676, -0.3162277660168379, -0.5, 
		-0.316227766016838, 0.3162277660168379, 0.5, 0.316227766016838};
  
  double vy []={0., 0., 0., 0.547722557505166, 0.8660254037844386, 0.5477225575051662, 
 0.547722557505166, 0.8660254037844386, 0.5477225575051662, 0., 0., 0., 
 -0.547722557505166, -0.8660254037844386, -0.5477225575051662, 
 -0.547722557505166, -0.8660254037844386, -0.5477225575051662};

  
  double vz []={-0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
 0.7745966692414833, -0.7745966692414834, 0., 0.7745966692414833, 
 -0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
		0.7745966692414833, -0.7745966692414834, 0., 0.7745966692414833};

  double ws [] ={3.4906585039886577, 5.585053606381854, 3.490658503988659, 
 3.4906585039886577, 5.585053606381854, 3.490658503988659, 
 3.4906585039886577, 5.585053606381854, 3.490658503988659, 
 3.4906585039886577, 5.585053606381854, 3.490658503988659, 
 3.4906585039886577, 5.585053606381854, 3.490658503988659, 
		 3.4906585039886577, 5.585053606381854, 3.490658503988659};
    
#endif

#ifdef D3Q1_6_3_plus

  //for z=1.0

  double vx []={0.6324555320336758, 1., 0.632455532033676, 0.3162277660168379, 0.5, 
 0.316227766016838, -0.3162277660168379, -0.5, -0.316227766016838, 
 -0.6324555320336758, -1., -0.632455532033676, -0.3162277660168379, -0.5, 
		-0.316227766016838, 0.3162277660168379, 0.5, 0.316227766016838,0};
  
  double vy []={0., 0., 0., 0.547722557505166, 0.8660254037844386, 0.5477225575051662, 
 0.547722557505166, 0.8660254037844386, 0.5477225575051662, 0., 0., 0., 
 -0.547722557505166, -0.8660254037844386, -0.5477225575051662, 
		-0.547722557505166, -0.8660254037844386, -0.5477225575051662,0};

  
  double vz []={-0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
 0.7745966692414833, -0.7745966692414834, 0., 0.7745966692414833, 
 -0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
		0.7745966692414833, -0.7745966692414834, 0., 0.7745966692414833,0};

  double ws [] ={3.4906585039886577, 5.585053606381854, 3.490658503988659, 
 3.4906585039886577, 5.585053606381854, 3.490658503988659, 
 3.4906585039886577, 5.585053606381854, 3.490658503988659, 
 3.4906585039886577, 5.585053606381854, 3.490658503988659, 
 3.4906585039886577, 5.585053606381854, 3.490658503988659, 
		 3.4906585039886577, 5.585053606381854, 3.490658503988659,75.39822368615503};
#endif

#ifdef D3Q1_8_3

  //for z=1.0

  double vx []={0.6324555320336758, 1., 0.632455532033676, 0.4472135954999578, 
 0.7071067811865475, 0.447213595499958, 0., 0., 0., -0.4472135954999578, 
 -0.7071067811865475, -0.447213595499958, -0.6324555320336758, -1., 
 -0.632455532033676, -0.4472135954999578, -0.7071067811865475, 
 -0.447213595499958, 0., 0., 0., 0.4472135954999578, 0.7071067811865475, 
		0.447213595499958};
  
  double vy []={0., 0., 0., 0.4472135954999578, 0.7071067811865475, 0.447213595499958, 
 0.6324555320336758, 1., 0.632455532033676, 0.4472135954999578, 
 0.7071067811865475, 0.447213595499958, 0., 0., 0., -0.4472135954999578, 
 -0.7071067811865475, -0.447213595499958, -0.6324555320336758, -1., 
 -0.632455532033676, -0.4472135954999578, -0.7071067811865475, 
		-0.447213595499958};


  
  double vz []={-0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
 0.7745966692414833, -0.7745966692414834, 0., 0.7745966692414833, 
 -0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
 0.7745966692414833, -0.7745966692414834, 0., 0.7745966692414833, 
 -0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
		0.7745966692414833};


  double ws [] ={2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
		 2.6179938779914935, 4.1887902047863905, 2.617993877991495};

#endif

#ifdef D3Q1_8_3_plus

  //for z=1.0

  double vx []={0.6324555320336758, 1., 0.632455532033676, 0.4472135954999578, 
 0.7071067811865475, 0.447213595499958, 0., 0., 0., -0.4472135954999578, 
 -0.7071067811865475, -0.447213595499958, -0.6324555320336758, -1., 
 -0.632455532033676, -0.4472135954999578, -0.7071067811865475, 
 -0.447213595499958, 0., 0., 0., 0.4472135954999578, 0.7071067811865475, 
		0.447213595499958,0.};
  
  double vy []={0., 0., 0., 0.4472135954999578, 0.7071067811865475, 0.447213595499958, 
 0.6324555320336758, 1., 0.632455532033676, 0.4472135954999578, 
 0.7071067811865475, 0.447213595499958, 0., 0., 0., -0.4472135954999578, 
 -0.7071067811865475, -0.447213595499958, -0.6324555320336758, -1., 
 -0.632455532033676, -0.4472135954999578, -0.7071067811865475, 
		-0.447213595499958,0.};


  
  double vz []={-0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
 0.7745966692414833, -0.7745966692414834, 0., 0.7745966692414833, 
 -0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
 0.7745966692414833, -0.7745966692414834, 0., 0.7745966692414833, 
 -0.7745966692414834, 0., 0.7745966692414833, -0.7745966692414834, 0., 
		0.7745966692414833,0.};


  double ws [] ={2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
 2.6179938779914935, 4.1887902047863905, 2.617993877991495, 
		 2.6179938779914935, 4.1887902047863905, 2.617993877991495,75.39822368615503};
#endif

#ifdef D3Q1_8_4_plus

  //for z=1.0

  double vx []={0.5083741268536301, 0.9404322888985427, 0.9404322888985427, 
 0.5083741268536301, 0.35947479247799197, 0.6649860487269458, 
 0.6649860487269458, 0.35947479247799197, 0., 0., 0., 0., 
 -0.35947479247799197, -0.6649860487269458, -0.6649860487269458, 
 -0.35947479247799197, -0.5083741268536301, -0.9404322888985427, 
 -0.9404322888985427, -0.5083741268536301, -0.35947479247799197, 
 -0.6649860487269458, -0.6649860487269458, -0.35947479247799197, 0., 0., 0., 
 0., 0.35947479247799197, 0.6649860487269458, 0.6649860487269458, 
		0.35947479247799197,0.};
  
  double vy []={0., 0., 0., 0., 0.35947479247799197, 0.6649860487269458, 0.6649860487269458, 
 0.35947479247799197, 0.5083741268536301, 0.9404322888985427, 
 0.9404322888985427, 0.5083741268536301, 0.35947479247799197, 
 0.6649860487269458, 0.6649860487269458, 0.35947479247799197, 0., 0., 0., 0., 
 -0.35947479247799197, -0.6649860487269458, -0.6649860487269458, 
 -0.35947479247799197, -0.5083741268536301, -0.9404322888985427, 
 -0.9404322888985427, -0.5083741268536301, -0.35947479247799197, 
		-0.6649860487269458, -0.6649860487269458, -0.35947479247799197,0.};
  
  double vz []={-0.8611363115940526, -0.33998104358485626, 0.33998104358485626, 
 0.8611363115940526, -0.8611363115940526, -0.33998104358485626, 
 0.33998104358485626, 0.8611363115940526, -0.8611363115940526, 
 -0.33998104358485626, 0.33998104358485626, 0.8611363115940526, 
 -0.8611363115940526, -0.33998104358485626, 0.33998104358485626, 
 0.8611363115940526, -0.8611363115940526, -0.33998104358485626, 
 0.33998104358485626, 0.8611363115940526, -0.8611363115940526, 
 -0.33998104358485626, 0.33998104358485626, 0.8611363115940526, 
 -0.8611363115940526, -0.33998104358485626, 0.33998104358485626, 
 0.8611363115940526, -0.8611363115940526, -0.33998104358485626, 
		0.33998104358485626, 0.8611363115940526,0.};


  double ws [] ={1.639227338999161, 3.0731616413855294, 3.0731616413855294, 
 1.639227338999161, 1.639227338999161, 3.0731616413855294, 
 3.0731616413855294, 1.639227338999161, 1.639227338999161, 
 3.0731616413855294, 3.0731616413855294, 1.639227338999161, 
 1.639227338999161, 3.0731616413855294, 3.0731616413855294, 
 1.639227338999161, 1.639227338999161, 3.0731616413855294, 
 3.0731616413855294, 1.639227338999161, 1.639227338999161, 
 3.0731616413855294, 3.0731616413855294, 1.639227338999161, 
 1.639227338999161, 3.0731616413855294, 3.0731616413855294, 
 1.639227338999161, 1.639227338999161, 3.0731616413855294, 
		 3.0731616413855294, 1.639227338999161, 75.39822368615503};
#endif

#ifdef D3Q1_8_5_plus

  //for z=1.0

  double vx []={0.4228925239521018, 0.8426451222634056, 1., 0.8426451222634056, 
 0.4228925239521018, 0.2990301713996256, 0.5958400800862215, 
 0.7071067811865475, 0.5958400800862215, 0.2990301713996256, 0., 0., 0., 0., 
 0., -0.2990301713996256, -0.5958400800862215, -0.7071067811865475, 
 -0.5958400800862215, -0.2990301713996256, -0.4228925239521018, 
 -0.8426451222634056, -1., -0.8426451222634056, -0.4228925239521018, 
 -0.2990301713996256, -0.5958400800862215, -0.7071067811865475, 
 -0.5958400800862215, -0.2990301713996256, 0., 0., 0., 0., 0., 
 0.2990301713996256, 0.5958400800862215, 0.7071067811865475, 
		0.5958400800862215, 0.2990301713996256,0.};
  
  double vy []={0., 0., 0., 0., 0., 0.2990301713996256, 0.5958400800862215, 
 0.7071067811865475, 0.5958400800862215, 0.2990301713996256, 
 0.4228925239521018, 0.8426451222634056, 1., 0.8426451222634056, 
 0.4228925239521018, 0.2990301713996256, 0.5958400800862215, 
 0.7071067811865475, 0.5958400800862215, 0.2990301713996256, 0., 0., 0., 0., 
 0., -0.2990301713996256, -0.5958400800862215, -0.7071067811865475, 
 -0.5958400800862215, -0.2990301713996256, -0.4228925239521018, 
 -0.8426451222634056, -1., -0.8426451222634056, -0.4228925239521018, 
 -0.2990301713996256, -0.5958400800862215, -0.7071067811865475, 
		-0.5958400800862215, -0.2990301713996256,0.};
  
  double vz []={-0.9061798459386641, -0.538469310105683, 0., 0.5384693101056831, 
 0.9061798459386641, -0.9061798459386641, -0.538469310105683, 0., 
 0.5384693101056831, 0.9061798459386641, -0.9061798459386641, 
 -0.538469310105683, 0., 0.5384693101056831, 0.9061798459386641, 
 -0.9061798459386641, -0.538469310105683, 0., 0.5384693101056831, 
 0.9061798459386641, -0.9061798459386641, -0.538469310105683, 0., 
 0.5384693101056831, 0.9061798459386641, -0.9061798459386641, 
 -0.538469310105683, 0., 0.5384693101056831, 0.9061798459386641, 
 -0.9061798459386641, -0.538469310105683, 0., 0.5384693101056831, 
 0.9061798459386641, -0.9061798459386641, -0.538469310105683, 0., 
		0.5384693101056831, 0.9061798459386641,0.};


  double ws [] ={1.116491642295654, 2.2554844725573897, 2.68082573106329, 2.255484472557388, 
 1.116491642295654, 1.116491642295654, 2.2554844725573897, 2.68082573106329, 
 2.255484472557388, 1.116491642295654, 1.116491642295654, 2.2554844725573897, 
 2.68082573106329, 2.255484472557388, 1.116491642295654, 1.116491642295654, 
 2.2554844725573897, 2.68082573106329, 2.255484472557388, 1.116491642295654, 
 1.116491642295654, 2.2554844725573897, 2.68082573106329, 2.255484472557388, 
 1.116491642295654, 1.116491642295654, 2.2554844725573897, 2.68082573106329, 
 2.255484472557388, 1.116491642295654, 1.116491642295654, 2.2554844725573897, 
 2.68082573106329, 2.255484472557388, 1.116491642295654, 1.116491642295654, 
		 2.2554844725573897, 2.68082573106329, 2.255484472557388, 1.116491642295654, 75.39822368615503};
#endif

#ifdef D3Q18

  //for z=1.0

  double vx []={1., -1., 0., 0., 0., 0., 0.7071067811865475, 0.7071067811865475, 
 0.7071067811865475, 0.7071067811865475, -0.7071067811865475, 
 -0.7071067811865475, -0.7071067811865475, -0.7071067811865475, 0., 0., 0., 
		0.};

  
  double vy []={0., 0., 1., -1., 0., 0., 0.7071067811865475, -0.7071067811865475, 0., 0., 
 0.7071067811865475, -0.7071067811865475, 0., 0., 0.7071067811865475, 
		0.7071067811865475, -0.7071067811865475, -0.7071067811865475};

  
  double vz []={0., 0., 0., 0., 1., -1., 0., 0., 0.7071067811865475, -0.7071067811865475, 
 0., 0., 0.7071067811865475, -0.7071067811865475, 0.7071067811865475, 
		-0.7071067811865475, 0.7071067811865475, -0.7071067811865475};

  double ws [] ={-25.132741228718345, -25.132741228718345, -25.132741228718345, 
 -25.132741228718345, -25.132741228718345, -25.132741228718345, 
 18.84955592153876, 18.84955592153876, 18.84955592153876, 18.84955592153876, 
 18.84955592153876, 18.84955592153876, 18.84955592153876, 18.84955592153876, 
		 18.84955592153876, 18.84955592153876, 18.84955592153876, 18.84955592153876};
#endif

#ifdef D3Q26

  //for z=1.0

  double vx []={-1., 0., 0., -0.7071067811865475, -0.7071067811865475, -0.7071067811865475, 
 -0.7071067811865475, 0., 0., -0.5773502691896258, -0.5773502691896258, 
 -0.5773502691896258, -0.5773502691896258, 1., 0., 0., 0.7071067811865475, 
 0.7071067811865475, 0.7071067811865475, 0.7071067811865475, 0., 0., 
 0.5773502691896258, 0.5773502691896258, 0.5773502691896258, 
		0.5773502691896258};


  
  double vy []={0., -1., 0., -0.7071067811865475, 0.7071067811865475, 0., 0., 
 -0.7071067811865475, -0.7071067811865475, -0.5773502691896258, 
 -0.5773502691896258, 0.5773502691896258, 0.5773502691896258, 0., 1., 0., 
 0.7071067811865475, -0.7071067811865475, 0., 0., 0.7071067811865475, 
 0.7071067811865475, 0.5773502691896258, 0.5773502691896258, 
		-0.5773502691896258, -0.5773502691896258};

  
  double vz []={0., 0., -1., 0., 0., -0.7071067811865475, 0.7071067811865475, 
 -0.7071067811865475, 0.7071067811865475, -0.5773502691896258, 
 0.5773502691896258, -0.5773502691896258, 0.5773502691896258, 0., 0., 1., 0., 
 0., 0.7071067811865475, -0.7071067811865475, 0.7071067811865475, 
 -0.7071067811865475, 0.5773502691896258, -0.5773502691896258, 
		0.5773502691896258, -0.5773502691896258};
  
  double w1=1.;

  double ws [] ={w1, w1, w1, 10.053096491487338 - 2.*w1, 10.053096491487338 - 2.*w1, 
 10.053096491487338 - 2.*w1, 10.053096491487338 - 2.*w1, 
 10.053096491487338 - 2.*w1, 10.053096491487338 - 2.*w1, 
 9.42477796076938 - 1.5*(10.053096491487338 - 2.*w1) - 0.75*w1, 
 9.42477796076938 - 1.5*(10.053096491487338 - 2.*w1) - 0.75*w1, 
 9.42477796076938 - 1.5*(10.053096491487338 - 2.*w1) - 0.75*w1, 
 9.42477796076938 - 1.5*(10.053096491487338 - 2.*w1) - 0.75*w1, w1, w1, w1, 
 10.053096491487338 - 2.*w1, 10.053096491487338 - 2.*w1, 
 10.053096491487338 - 2.*w1, 10.053096491487338 - 2.*w1, 
 10.053096491487338 - 2.*w1, 10.053096491487338 - 2.*w1, 
 9.42477796076938 - 1.5*(10.053096491487338 - 2.*w1) - 0.75*w1, 
 9.42477796076938 - 1.5*(10.053096491487338 - 2.*w1) - 0.75*w1, 
 9.42477796076938 - 1.5*(10.053096491487338 - 2.*w1) - 0.75*w1, 
		 9.42477796076938 - 1.5*(10.053096491487338 - 2.*w1) - 0.75*w1};
#endif

  for (int i=0;i<SPEEDS;i++)
    {
      c[1][i]=vx[i];
      c[2][i]=vy[i];
      c[3][i]=vz[i];
      c[0][i]=sqrt(c[1][i]*c[1][i]+c[2][i]*c[2][i]+c[3][i]*c[3][i]);
#ifdef plus
      c[0][SPEEDS-1]=1.;
#endif
      w[i] = ws[i]*exp(REFZ*c[0][i])*dof*M_PI/1440.; 
      for (int a=0;a<1+CDIM;a++)
	for (int b=0;b<1+CDIM;b++)
	  {
	    wtw[i][a][b]=w[i]*c[a][i]*c[b][i];
	  }

      //need fix for diagonals for alpha!=1
#ifdef plus      
      for (int a=1;a<4;a++)
	wtw[i][a][a]=w[i]*(c[a][i]*c[a][i]-(c[1][i]*c[1][i]+c[2][i]*c[2][i]+c[3][i]*c[3][i]-1)/3);
#endif
    }	
  
 
  for (int i=0;i<SPEEDS;i++)
    {
      poly[0][i]=(c[1][i]*c[1][i]+c[2][i]*c[2][i]+c[3][i]*c[3][i]);
      poly[1][i]=c[1][i];
      poly[2][i]=c[2][i];
      poly[3][i]=c[3][i];
      poly[4][i]=c[1][i]*c[1][i]-1/3.*(c[1][i]*c[1][i]+c[2][i]*c[2][i]+c[3][i]*c[3][i]);
      poly[5][i]=c[1][i]*c[2][i];
      poly[6][i]=c[1][i]*c[3][i];
      poly[7][i]=c[2][i]*c[2][i]-1/3.*(c[1][i]*c[1][i]+c[2][i]*c[2][i]+c[3][i]*c[3][i])+0.5*poly[4][i];
      poly[8][i]=c[2][i]*c[3][i];

#ifdef plus
      poly[9][i]=1-(c[1][i]*c[1][i]+c[2][i]*c[2][i]+c[3][i]*c[3][i]); 
      poly[10][i]=c[1][i]*c[1][i]*c[1][i]-3./5.*c[1][i]*(c[1][i]*c[1][i]+c[2][i]*c[2][i]+c[3][i]*c[3][i]); //only D3S8-5-plus
#endif
      
      
#ifdef MILNE
      dzpoly[0][i]=0;
      dzpoly[1][i]=0;
      dzpoly[2][i]=0;
      dzpoly[3][i]=1;
      dzpoly[4][i]=0;
      dzpoly[5][i]=0;
      dzpoly[6][i]=c[1][i];
      dzpoly[7][i]=0;
      dzpoly[8][i]=c[2][i];
#endif
    }
  

  

  for (int n=0;n<numNpoly;n++)
    {
      Npoly[n]=0;
      for (int i=0;i<SPEEDS;i++)
	{
	  Npoly[n]+=ws[i]*poly[n][i]*poly[n][i];
	}
      //printf("npoly %i %f\n",n,Npoly[n]);
    }

  
  coord co;
  for (int j=0;j<SPEEDS;j++)
    {
      co.set(mysign(-c[1][j]/c[0][j]),mysign(-c[2][j]/c[0][j]),mysign(-c[3][j]/c[0][j]));
      dist[j]=co.veclen();//calculate distance between NN and reference point
      //printf("j=%i c=%g %g %g\n",j,c[1][j]/c[0][j],c[2][j]/c[0][j],c[3][j]/c[0][j]);
    }

  printf("finished\n");
}

#ifdef MILNE
void setdfdz1(xygrid *df,xygrid *localf,int locald)
{

  //via (M1)

  double arr[numNpoly];
  
  if (locald)
    {
      
      arr[0]=0;
      for (int i=0;i<SPEEDS;i++)
	arr[0]+=w[i]*exp(-1/1.1*c[0][i])*c[0][i]*c[0][i];    
      //printf("This is setdfdz; sanity %g\n",arr[0]);
    }
  

  for (int n=0;n<numNpoly;n++)
    {
      arr[n]=0;
      for (int i=0;i<SPEEDS;i++)
	{
	  arr[n]-=w[i]*localf->data[i]*(poly[n][i]*(1+c[3][i]*c[3][i]/c[0][i]/c[0][i])+c[3][i]*dzpoly[n][i]);	  
	}
      
      //if (locald)
      // 	printf("n=%i a=%g\n",n,arr[n]);

      arr[n]/=Npoly[n];
       

    }
      
  for (int i=0;i<SPEEDS;i++)
    {
      df->data[i]=0;
      for (int n=0;n<numNpoly;n++)      
	{
	  df->data[i]+=arr[n]*poly[n][i];
	}
      df->data[i]*=exp(-REFZ*c[0][i]);
      // if (locald)
      //	printf("i=%i f(i)=%g\n",i,df->data[i]);
    }  
    
}


#endif

void setfTab(xygrid *feq,double *tabini,int locald)
{
  if (locald)
    printf("f_ini setting T00=%f\n",tabini[0]);
  double arr[numNpoly];
  
  double cdof=24./(M_PI/60.*dof);

  arr[0]=tabini[0]*cdof/Npoly[0];
  arr[1]=tabini[1]*cdof/Npoly[1];
  arr[2]=tabini[2]*cdof/Npoly[2];
  arr[3]=tabini[3]*cdof/Npoly[3];
  arr[4]=(tabini[4]-tabini[0]/3.)*cdof/Npoly[4];
  arr[5]=tabini[5]*cdof/Npoly[5];
  arr[6]=tabini[6]*cdof/Npoly[6];
  arr[7]=(tabini[7]-tabini[0]/3.+0.5*(tabini[4]-tabini[0]/3.))*cdof/Npoly[7];
  arr[8]=tabini[8]*cdof/Npoly[7];
  for (int j=9;j<numNpoly;j++)
    arr[j]=0.;
  
  for (int i=0;i<SPEEDS;i++)
    {
      double expweight=exp(-1.0*c[0][i]);

      feq->data[i]=0;
      for (int m=0;m<numNpoly;m++)    
	feq->data[i]+=expweight*(arr[m]*poly[m][i]);
    }
  if (locald)
    {
      printf("array\n");
      printNvector(arr,numNpoly,0);
    }

  
}

void setfeq(xygrid *feq,double locT,double* v,double u0,int locald)
{
  if (locald)
    printf("\t debug feq setting T=%f v(0)=%f v(1)=%f v(2)=%f gamma=%f\n",locT,v[0],v[1],v[2],u0);



#ifndef feqsimple
 
  double arr[numNpoly];

  double u02=u0*u0;
  double v2=v[0]*v[0]+v[1]*v[1]+v[2]*v[2];
  double vabs=sqrt(v2);
  double u=u0*sqrt(v2);
  double T4=pow(locT,4);


  arr[0]=(1+4./3.*v2*u02)*T4;
  arr[1]=4*u02*v[0]*T4;
  arr[2]=4*u02*v[1]*T4;
  arr[3]=4*u02*v[2]*T4;
  arr[4]=15*u02*(v[0]*v[0]-v2/3.)*T4;
  arr[5]=20*u02*(v[0]*v[1])*T4;
  arr[6]=20*u02*(v[0]*v[2])*T4;
  arr[7]=4./3.*(15*u02*(v[1]*v[1]-v2/3.)*T4+0.5*arr[4]);
  arr[8]=20*u02*(v[1]*v[2])*T4;
#ifdef plus
  arr[0]=(1+4./3.*v2*u02)*T4+alpha;
  arr[9]=-alpha;
  arr[10]=a10*T4*v[0]*u0;
#endif

  
  //for (int j=numNpoly;j<SPEEDS;j++)
  //	arr[j]=0;

  
  /*
#ifdef D3Q1_6_3
  if (vabs>0.4)
    {
      arr[9]=T4*u0*v[0]*(c9+u02*(v[0]*v[0]*c9b+v[1]*v[1]*c9c));
      arr[10]=T4*u0*v[1]*(c10+u02*(v[0]*v[0]*c10b+v[1]*v[1]*c10c));
      arr[11]=0;
      arr[12]=T4*u0*v[0]*(c12+u02*(v[0]*v[0]*c12b+v[1]*v[1]*c12c));
      arr[13]=0;
      arr[14]=T4*(c14+u02*(v[0]*v[0]*c14b+v[1]*v[1]*c14c));
      arr[15]=T4*u02*v[0]*v[1]*c15;
      arr[16]=0;
      arr[17]=T4*u0*v[0]*(c17+u02*(v[0]*v[0]*c17b+v[1]*v[1]*c17c));
    }
  else
    {
      for (int j=9;j<SPEEDS;j++)
	arr[j]=0;
    }
    #endif*/
  
  for (int i=0;i<SPEEDS;i++)
    {
      double expweight=exp(-1.0*c[0][i]);
      
      //arr[9]=T4*175./480.*u02*u0*(3*v[0]*v2-5*v[0]*v[0]*v[0])*(-4*u*u0*(15-10*u*u-8*u*u*u*u)+60*atan(u/u0));
      //arr[10]=-15*u02*v[1]*T4;
      //arr[11]=-15*u02*v[2]*T4;

      feq->data[i]=0;
      for (int m=0;m<numNpoly;m++)    
	feq->data[i]+=expweight*(arr[m]*poly[m][i]);
      if ((feq->data[i]<0)&&(i!=SPEEDS-1))
	{
	  if (!NEGFEQWARN)
	    {
	      printf("trouble brewing at i=%i func=%g, T=%f vx=%f vy=%f vz=%f u=%f\n",i,feq->data[i],locT,v[0],v[1],v[2],sqrt(v2));
	      NEGFEQWARN=1;
	    }	  
	  else
	    {
	      //don't do anything
	      NEGFEQWARN=1;
	    }
	  //printf("array\n");
	  //printNvector(arr,numNpoly,0);	  
	  //printf("now 0=%g 1_0=%g 1_1=%g 2_00=%g 2_01=%g\n",a0_0,a1_0*poly[1][i],a1_1*poly[2][i],a2_00*poly[3][i],a2_01*poly[4][i]);
	}
    }
  if (locald)
	{
	  printf("array\n");
	  printNvector(arr,numNpoly,0);
	}

  // printf("checking feq: z=%g a0=%g a1=%g a5=%g\n",z,arr[0],arr[1],arr[5]);
  //double sum=0;
  //for (int i=0;i<SPEEDS;i++)
  // sum+=feq->data[i]*w[i];
  //printf("z=%g checking a0 %g\n",z,sum/Npoly[0]);
#endif
  
}

//inverse square root of minus metric g33
double sg33(double x)
{
  double res=1.;
#ifdef MILNE
  res=1./(x*AT);
#endif
  return res;
}

void initialize(xygrid * XCgrid)
{

  printf("===> Info: Initializing space grid ");

#ifdef INITGAUSSIAN
  printf(" from Gaussian temperature distribution...\t");
#endif

#ifdef BAYMIC
  printf(" from Box temperature distribution...\t");
#endif

#ifdef INITHANS
  fstream inited;
  printf(" from energy density in interpx.dat file...\t");
  inited.open("init/interpx.dat", ios::in);
#endif

#ifdef INITFILE
  printf(" from energy density in inited.dat file...\t");
  fstream inited;
  if (NX==139)
    inited.open("init/inited-139.dat", ios::in);
  if (NX==199)
    inited.open("init/inited-199.dat", ios::in);
  if (NX==279)
    inited.open("init/inited-279.dat", ios::in);
  if (NX==559)
    inited.open("init/inited-559.dat", ios::in);
  if (NX==1119)
    inited.open("init/inited-1119.dat", ios::in);
#endif

  coord co;
  double localT;
  double ed;
  double myu0 [] = {0.,0.,0.,0.};
  double dummyu0val=1.;

  double stat[4];

  xygrid feq;
  feq.init(SPEEDS,1,1);
  feq.alloc_data();
  
  double Tmin=0.01;

  
  int total=0;
  for(long int i=0;i<XCgrid->SIZE;i++)
    {
      XCgrid->getcoord(&co,i);
      //Gaussian
#ifdef INITGAUSSIAN
      localT=regulator+0.5*exp(-pow(AT/fmtoGeVI*(co.x-XCgrid->Middle.x),2)-pow(AT/fmtoGeVI*(co.y-XCgrid->Middle.y),2)-pow(AT/fmtoGeVI*(co.z-XCgrid->Middle.z),2));
#endif

#ifdef BAYMIC
      double rr;
      rr=sqrt(pow(AT/fmtoGeVI*(co.x-XCgrid->Middle.x),2)+pow(AT/fmtoGeVI*(co.y-XCgrid->Middle.y),2)+pow(AT/fmtoGeVI*(co.z-XCgrid->Middle.z),2));
      if (rr<6.4)
	localT=1.0;
      else
	localT=regulator;
#endif

#ifdef BAYMMODIC
      double rr;
      rr=sqrt(pow(AT/fmtoGeVI*(co.x-XCgrid->Middle.x),2)+pow(AT/fmtoGeVI*(co.y-XCgrid->Middle.y),2)+pow(AT/fmtoGeVI*(co.z-XCgrid->Middle.z),2));
      localT=regulator+(1.0-regulator)/(1+exp((rr-6.4)/regsigma));
      
#endif



#ifdef FS1D
      double regsigma=0.01;
      /*
      if (AT/fmtoGeVI*(co.x-XCgrid->Middle.x)<=0.)
	localT=1.0;
      else
      localT=regulator;*/
      localT=regulator+(1.-regulator)/(1.+exp((co.x-XCgrid->Middle.x)*AT/fmtoGeVI/regsigma));
      
#endif

#ifdef INITFILE
      if(inited.is_open())
	{
	  inited >> ed;
	  localT=Tofed(ed/pow(AT,4));
	  //printf("localT=%g\n",localT);
	}
      else
	{
	  printf("\nERROR: file inited.dat could not be opened\n");
	  ABORT=1;
	  i=XCgrid->SIZE-1;
	}
#endif
      
#ifdef INITHANS
      double dummy,du0,dux,duy,duz;
      double dumtab[9];

      if(inited.is_open())
	{
	  //fgets(dbuffer,255,inited);	  
	  inited >> dummy;
	  inited >> dummy;
	  inited >> dummy;
	  inited >> ed;
	  inited >> du0;
	  inited >> dux;
	  inited >> duy;
	  inited >> duz;
	  
	  for (int ib=0;ib<10;ib++)
	    {
	      inited >> dumtab[ib]; //tab components
	      dumtab[ib]*=pow(fmtoGeVI,4);
	    }
	  
	  
	  
	  //we're assuming that energy scale is fm^-1, 
	  //so ed needs to be rescaled accordingly
	  ed*=pow(fmtoGeVI,4);
	  
	  

	  localT=Tofed(ed);
	  dummyu0val=sqrt(1+dux*dux+duy*duy+duz*duz);
	  
	  myu0[0]=dux/dummyu0val;
	  myu0[1]=duy/dummyu0val;
	  myu0[2]=duz/dummyu0val;

	  //if (i==XCgrid->intMiddle)
	  //  printf("middle: ed=%f T=%f t00=%f\n t01=%f",ed,localT,dumtab[0],dumtab[1]);

	  //myu0[0]=dux;
	  //myu0[1]=duy;
	  //myu0[2]=duz;
	  
	  //printf("localT=%g\n",localT);
	  total++;
	}
      else
	{
	  printf("\nERROR: file inited.dat could not be opened\n");
	  
	  ABORT=1;
	  i=XCgrid->SIZE-1;
	}
      
#endif
      
      debug=0;
      /*if ((co.x==XCgrid->Middle.x)&&(co.y==XCgrid->Middle.y+65))
	  debug=1;
	if ((co.x==XCgrid->Middle.x+65)&&(co.y==XCgrid->Middle.y))
	debug=1;*/
#ifndef INITHANS
	setfeq(&feq,localT,myu0,dummyu0val,debug);
#else
	setfeq(&feq,localT,myu0,dummyu0val,debug);
	//setfTab(&feq,dumtab,debug);
#endif

      for (int j=0;j<SPEEDS;j++)
	{
	  f[i].data[j]=feq.data[j];
	  fnew[i].data[j]=f[i].data[j];
	}
      calcEMT(Tab[i].data,&f[i],0.);

      if (i==XCgrid->intMiddle)
	    {
	      printf("\n");
	      printNxNmatrix(Tab[i].data,4,1);
	    }
      //calcedvnew(Tab[i].data,stat);
      //if (i==XCgrid->intMiddle)
      //	printf("Norm: stat[0]=%f\n",stat[0]);
      calcedvquick(Tab[i].data,stat,0);
      /*if (i==XCgrid->intMiddle)
	{
	  calcedvquick(Tab[i].data,stat,0);
	  for (int m=0;m<4;m++)
	    printf("quick: stat[%i]=%f\n",m,stat[m]);
	    }*/
      for (int a=0;a<CDIM;a++)
	{
	  v[i].data[a]=stat[a+1];						
	}
      if (debug)
	{
	  printf("got stat[0]=%g (%g,%g,%g)\n",stat[0],stat[1],stat[2],stat[3]);
	  /*printf("what's going on?\n");
	  double testme=0;
	  for (int i=0;i<SPEEDS;i++)
	  {	    
	    testme+=wtw[i][a][]*pf->data[i];
	    }*/
	}

      T[i].data[0]=Tofed(stat[0]);
      u0[i].data[0]=sqrt(1./(1-v[i].data[0]*v[i].data[0]-v[i].data[1]*v[i].data[1]-v[i].data[2]*v[i].data[2]));
      warn[i]=0;

      //      if ((co.x==XCgrid->Middle.x)&&(co.y==XCgrid->Middle.y)&&(co.z==XCgrid->Middle.z)) 
      //printf("init center T=%g\n",localT);
      
      // if (i==XCgrid->intMiddle)
      //	printf("middle: ed=%f T=%f t00=%f\n t01=%f",stat[0],T[i].data[0],Tab[i].data[0],Tab[i].data[1]);


    }
  


#ifdef INITFILE
  inited.close();
#endif
#ifdef INITHANS
  inited.close();  
  
#endif


  if (!ABORT)
    printf("finished\n");

  printf("===> Info: read a total of %i points from data file\n",total);
}

void calcEMT(double * res,xygrid *pf,double regulator)
{

  double tab[CDIM+1][CDIM+1];
  int msign=1;
  double mod;

  for (int a=0;a<1+CDIM;a++)
    for (int b=0;b<=a;b++)
      {
	msign=1;
	tab[a][b]=0;
	for (int i=0;i<SPEEDS;i++)
	  {	    
	    tab[a][b]+=wtw[i][a][b]*pf->data[i];
	  }							
	if ((a==b)&&(b==0))
	  tab[a][b]+=regulator;
	if ((a==b)&&(b>0))
	tab[a][b]+=regulator/3.;
	if (b>0)
	  msign*=-1;
	res[a*(1+CDIM)+b]=msign*tab[a][b];
	if ((b<a)&&(b>0))
	  res[b*(1+CDIM)+a]=res[a*(1+CDIM)+b];
	if ((b==0)&&(a>0))
	  res[b*(1+CDIM)+a]=-res[a*(1+CDIM)+b];
      }  
}



double Tofed(double ed)
{
  double res=0;
  if (ed>exp(logeot.low))
    res=logeot.f(log(ed));
  else
    res=logeot.f(logeot.low);
  return res;
}

double Tofed(double ed,gsl_interp_accel *acc)
{
  double res=0;
  if (ed>exp(logeot.low))
    res=logeot.f(log(ed),acc);
  else
    res=logeot.f(logeot.low,acc);
  return res;
}


int calcedvquick(double *tab, double *statevec,int mdebug)
{
  int WARN=0;
  double uguess[]={1,0,0,0};
  double uguessnew[]={0,0,0,0};
  double eu0=-1.0;
  int stepper=0;

  if (mdebug)
    printNxNmatrix(tab,4,1);

  while((fabs((uguessnew[0]-eu0)/eu0)>1.e-5)&&(stepper<100))
    {
      eu0=uguessnew[0];      
      for (int i=0;i<4;i++)
	uguessnew[i]=0.0;
      
      for (int i=0;i<4;i++)
	for (int j=0;j<4;j++)
	  uguessnew[i]+=tab[4*i+j]*uguess[j];

      for (int i=0;i<4;i++)
	{ 
	  if (mdebug)
	    printf("i=%i u=%g\n",i,uguessnew[i]);
	  uguess[i]=uguessnew[i]/uguessnew[0];
	}      
      stepper++;
      if (mdebug)
	printf("guess %f\n",uguessnew[0]);
    }
  
  if (isnan(uguessnew[0]))
    WARN=1;


  for (int a=1;a<4;a++)
    {
      if (isnan(uguess[a]))
	WARN=1;
      else
	statevec[a]=uguess[a];
    }

  double u2=statevec[1]*statevec[1]+statevec[2]*statevec[2]+statevec[3]*statevec[3];

  statevec[0]=uguessnew[0];

  if ((u2>0.99)||(statevec[0]<0.)||statevec[0]>exp(logeot.high))
    {
      /*
      for (int a=1;a<4;a++)
	statevec[a]/=sqrt(u2+0.001);
	statevec[0]=fabs(uguessnew[0]);*/
      WARN=1;
    }
  //statevec[4]=sqrt(fabs(1/(1-u2)));

  

  if (mdebug)
    printf("u2=%f 1=%f 2=%f 3=%f\n",u2,statevec[1],statevec[2],statevec[3]);
  return WARN;
}

/*
int calcedvnew(double *tab,double *statevec)
{
  
  int WARN=0;

  gsl_vector_complex *eval;
  gsl_matrix_complex *evec;
  gsl_matrix_view m;
  gsl_eigen_nonsymmv_workspace *WS;
  gsl_vector_view revec;
  gsl_vector_view reval;

  eval = gsl_vector_complex_alloc (4);
  evec = gsl_matrix_complex_alloc (4, 4);
  WS = gsl_eigen_nonsymmv_alloc (4);
  
  int a;

  double holdmat[16];

  for (a=0;a<16;a++)
    holdmat[a]=tab[a];

  m   = gsl_matrix_view_array (holdmat, 4, 4);
  
  int res=gsl_eigen_nonsymmv (&m.matrix, eval, evec, WS);

  if (res)
    {
      printf ("error in calcedvnew: %s\n",gsl_strerror (a));
      printNxNmatrix(tab,4,1);
    }

  //checking for imaginary parts
  reval = gsl_vector_complex_imag(eval);
  int maxval=0;
  while((maxval<4)&&(!WARN))
    {
      if (fabs(gsl_vector_get(&reval.vector,maxval))>1.e-10)
	{
	  if (OUTPUTWARN)
	    {
	      printf("WARNING found imaginary eigenvalues\n");
	      printNxNmatrix(tab,4,1);
	      printf("maxval=%i value=%g\n",maxval,gsl_vector_get(&reval.vector,maxval));
	    }
	  WARN=1; //warncode imaginary EV
	}
      maxval++;
    }

  reval = gsl_vector_complex_real(eval);
  


  holdmat[0]=-1;

  
  
  
  
  maxval=0;
  while(holdmat[0]<0)
    {	      
      if (maxval<4)
	{
	  revec = gsl_vector_complex_real(&(gsl_matrix_complex_column (evec, maxval).vector));
	  holdmat[0]=gsl_vector_get(&revec.vector,0)*gsl_vector_get(&revec.vector,0);
	  holdmat[0]-=gsl_vector_get(&revec.vector,1)*gsl_vector_get(&revec.vector,1);
	  holdmat[0]-=gsl_vector_get(&revec.vector,2)*gsl_vector_get(&revec.vector,2);
	  holdmat[0]-=gsl_vector_get(&revec.vector,3)*gsl_vector_get(&revec.vector,3);
	  maxval++;
	}
      else
	{
	  if (OUTPUTWARN)
	    {
	      printf("WARNING  couldn't find time-like eigenvector\n");
	      printf("Tab:\n");
	      printNxNmatrix(tab,4,1);
	    }
	  maxval=0;
	  revec = gsl_vector_complex_real(&(gsl_matrix_complex_column (evec, maxval).vector));
	  
	  holdmat[0]=1.;
	   maxval=1;
	  //ABORT=1; //try to rectify
	   if (WARN==0)
	     WARN=2; //no time-like EV
	}       
    }
  maxval--;
  
  statevec[0]=gsl_vector_get(&reval.vector,maxval);

  if (((statevec[0]<0)||(statevec[0]>exp(logeot.high)))&&(WARN==0))
    WARN=3; //eigenvalue out of range
  	    
  for (a=1;a<1+CDIM;a++)	           
    statevec[a]=gsl_vector_get(&revec.vector,a)/gsl_vector_get(&revec.vector,0);
 
  
  gsl_eigen_nonsymmv_free (WS);
  gsl_vector_complex_free (eval);
  gsl_matrix_complex_free (evec);

  return WARN;
  }*/

double edofT(double tempT)
{
  double myresult;
  myresult=M_PI*M_PI/60.*dof*pow(tempT,4);
  return myresult;
}

double comppi(long int i)
{
    
  double pixx=Tab[i].data[5]+4./3.*edofT(T[i].data[0])*pow(v[i].data[0]*u0[i].data[0],2)+edofT(T[i].data[0])/3.*(1+pow(v[i].data[0]*u0[i].data[0],2));
  double piyy=Tab[i].data[10]+4./3.*edofT(T[i].data[0])*pow(v[i].data[1]*u0[i].data[0],2)+edofT(T[i].data[0])/3.*(1+pow(v[i].data[1]*u0[i].data[0],2));
  double pizz=Tab[i].data[15]+4./3.*edofT(T[i].data[0])*pow(v[i].data[2]*u0[i].data[0],2)+edofT(T[i].data[0])/3.*(1+pow(v[i].data[2]*u0[i].data[0],2));
 
  
  return 0.33333*(fabs(pixx)+fabs(piyy)+fabs(pizz));
}

double compsigma(long int i,xygrid * XCgrid)
{
    
  
  
  coord co,conn;
  long int help;
  double dvx,dvy,dvz;
  XCgrid->getcoord(&co,i);
  conn.set(co.x+1,co.y,co.z);
  help=XCgrid->getind(&conn);
  dvx=0.5*v[help].data[0];
  conn.set(co.x-1,co.y,co.z);
  help=XCgrid->getind(&conn);
  dvx-=0.5*v[help].data[0];
  dvx/=2*AT;

  conn.set(co.x,co.y+1,co.z);
  help=XCgrid->getind(&conn);
  dvy=0.5*v[help].data[1];
  conn.set(co.x,co.y-1,co.z);
  help=XCgrid->getind(&conn);
  dvy-=0.5*v[help].data[1];
  dvy/=2*AT;

  conn.set(co.x,co.y,co.z+1);
  help=XCgrid->getind(&conn);
  dvz=0.5*v[help].data[2];
  conn.set(co.x,co.y,co.z-1);
  help=XCgrid->getind(&conn);
  dvz-=0.5*v[help].data[2];
  dvz/=2*AT;

  double sigmaxx=-(-4.*dvx+2*dvy+2*dvz)/3.;
  double sigmayy=-(-4.*dvy+2*dvx+2*dvz)/3.;
  double sigmazz=-(-4.*dvz+2*dvx+2*dvy)/3.;
  
  return 0.33333*(fabs(sigmaxx)+fabs(sigmayy)+fabs(sigmazz));
}

void output(xygrid *XCgrid)
{
  fstream outTx,outTy,outTz,outTxy,outTall;
  fstream outUx,outUy,outUz;
  fstream outwarnxy,outvabs;
  char buffer[255];

  int mycase=2;

  char outcase[2];

  if (mycase==1)
    sprintf(outcase,"%s","T");
    
  if (mycase==2)
    sprintf(outcase,"%s","E");

  sprintf(buffer,"%s/%s-x-%.2f.dat",outdir,outcase,t*AT/fmtoGeVI);
  outTx.open(buffer,ios::out); 

  //clear string
  buffer[0]='\0';

  if (mycase==1)
    sprintf(buffer,"#Temperature along x-axis\n #%s [fm]\t T [GeV]\n","x");
  if (mycase==2)
    sprintf(buffer,"#Energy density along x-axis\n #%s [fm]\t ED [GeV^4]\n","x");
  outTx.write(buffer,strlen(buffer));


  sprintf(buffer,"%s/%s-y-%.2f.dat",outdir,outcase,t*AT/fmtoGeVI);
  outTy.open(buffer,ios::out); 
  if (mycase==1)
    sprintf(buffer,"#Temperature along y-axis\n #x [fm]\t T [GeV]\n ");
  if (mycase==2)
    sprintf(buffer,"#Energy density along y-axis\n #x [fm]\t ED [GeV^4]\n ");
  outTy.write(buffer,strlen(buffer));
  

  sprintf(buffer,"%s/%s-z-%.2f.dat",outdir,outcase,t*AT/fmtoGeVI);
  outTz.open(buffer,ios::out); 
  if (mycase==1)
    sprintf(buffer,"#Temperature along z-axis\n #x [fm]\t T [GeV]\n ");
  if (mycase==2)
    sprintf(buffer,"#Energy density along z-axis\n #x [fm]\t ED [GeV^4]\n ");
  outTz.write(buffer,strlen(buffer));


  sprintf(buffer,"%s/U-x-%.2f.dat",outdir,t*AT/fmtoGeVI);
  outUx.open(buffer,ios::out); 
  sprintf(buffer,"#Velocity v^x=u^x/u^0 along x-axis\n #x [fm]\t v^x\n ");
  outUx.write(buffer,strlen(buffer));


  sprintf(buffer,"%s/U-y-%.2f.dat",outdir,t*AT/fmtoGeVI);
  outUy.open(buffer,ios::out); 
  sprintf(buffer,"#Velocity v^y=u^y/u^0 along y-axis\n #x [fm]\t v^y\n ");
  outUy.write(buffer,strlen(buffer));


  sprintf(buffer,"%s/U-z-%.2f.dat",outdir,t*AT/fmtoGeVI);
  outUz.open(buffer,ios::out); 
  sprintf(buffer,"#Velocity v^z=u^z/u^0 along z-axis\n #x [fm]\t v^z\n ");
  outUz.write(buffer,strlen(buffer));


  sprintf(buffer,"%s/%s-xy-%.2f.dat",outdir,outcase,t*AT/fmtoGeVI);
  outTxy.open(buffer,ios::out);
  if (mycase==1)
    sprintf(buffer,"#Temperature in x-y plane\n #x [fm]\t y [fm]\t\t T [GeV]\n ");
  if (mycase==2)
    sprintf(buffer,"#Energy Density in x-y plane\n #x [fm]\t y [fm]\t\t ED [GeV^4]\n ");
  outTxy.write(buffer,strlen(buffer));

  sprintf(buffer,"%s/warn-xy-%.2f.dat",outdir,t*AT/fmtoGeVI);
  outwarnxy.open(buffer,ios::out); 
  sprintf(buffer,"#warnings in xy plane\n ");
  outwarnxy.write(buffer,strlen(buffer));

  sprintf(buffer,"%s/v-r-%.2f.dat",outdir,t*AT/fmtoGeVI);
  outvabs.open(buffer,ios::out);
  sprintf(buffer,"#|v|       #r [fm]\t");
  outvabs.write(buffer,strlen(buffer));

  sprintf(buffer,"%s/%s-all-%.2f.dat",outdir,outcase,t*AT/fmtoGeVI);
  outTall.open(buffer,ios::out);
  if (mycase==1)
    sprintf(buffer,"#Temperature\n #x [fm]\t y [fm]\t z [fm]  \t\t T [GeV]\n ");
  if (mycase==2)
    sprintf(buffer,"#Energy density\n #x [fm]\t y [fm]\t z [fm]  \t\t ED [GeV^4]\n ");
  outTall.write(buffer,strlen(buffer));

  coord co;
  
  int help=0;
  for (long int i=0;i<XCgrid->SIZE;i++)
    help+=warn[i];
  
  //printf("total warn=%f\n",help);

  for (long int i=0;i<XCgrid->SIZE;i++)
    {

      XCgrid->getcoord(&co,i);
      //calcEMT(Tab[i].data,&f[i]);
      //      if (Tab[0]<0)
      //printf("trouble @ %f i=%li T00=%g\n",AT*(co.x-XCgrid->Middle.x),i,Tab[0]);
      //calcedv(Tab[i].data,stat);

      //sprintf(buffer, "%f\t  %f\t %f\t %g\n",AT*(co.x-XCgrid->Middle.x)/fmtoGeVI,AT*(co.y-XCgrid->Middle.y)/fmtoGeVI,AT*(co.z-XCgrid->Middle.z)/fmtoGeVI,T[i].data[0]);
      //outTall.write(buffer,strlen(buffer));

      double myresult;
      if (mycase==1)
	myresult=T[i].data[0];
      if (mycase==2)
	{
	  //double stat[4];
	  //calcedvnew(Tab[i].data,stat);
	  //myresult=Tab[i].data[0]-v[i].data[0]*Tab[i].data[1]-v[i].data[1]*Tab[i].data[2]-v[i].data[2]*Tab[i].data[2];
	  //myresult=stat[0];
	  myresult=M_PI*M_PI/60.*dof*pow(T[i].data[0],4)/pow(fmtoGeVI,4);
	}

      if ((co.z==XCgrid->Middle.z))
	{
	  sprintf(buffer, "%f\t %f\t %g\n",AT*(co.x-XCgrid->Middle.x)/fmtoGeVI,AT*(co.y-XCgrid->Middle.y)/fmtoGeVI,myresult);	  
	  outTxy.write(buffer,strlen(buffer));
	}  

      if ((co.z==XCgrid->Middle.z))
	{
	  sprintf(buffer, "%f\t %f\t %i\n",AT*(co.x-XCgrid->Middle.x)/fmtoGeVI,AT*(co.y-XCgrid->Middle.y)/fmtoGeVI,warn[i]);
	  outwarnxy.write(buffer,strlen(buffer));
	}  

      if (i==XCgrid->intMiddle)	
	{
	  //double ddd[3];
	  //compvisc(i,ddd);
	  printf("t=%f @ center T=%f ed=%f pi=%g sigma=%g\n",t*AT/fmtoGeVI,T[i].data[0],Tab[i].data[0],comppi(i),compsigma(i,XCgrid));
	  sprintf(buffer,"%f\t %f \t%g\t%g\n",t*AT/fmtoGeVI,Tab[i].data[0]*pow(fmtoGeVI,-4),comppi(i)*pow(fmtoGeVI,-4),compsigma(i,XCgrid)*pow(fmtoGeVI,-1));	  
	  meta.write(buffer,strlen(buffer));
	  meta.flush();
	  //printf("ed=%g\n",Tab[i].data[0]);
	}
     
      //printf("%i %f %f %f\n",i,Tab[0],Tab[1],Tab[2]);
      if ((co.y==XCgrid->Middle.y)&&(co.z==XCgrid->Middle.z))
	{
	  sprintf(buffer, "%f\t %g\n",AT*(co.x-XCgrid->Middle.x)/fmtoGeVI,myresult);
	  outTx.write(buffer,strlen(buffer));
	  sprintf(buffer, "%f\t %g\n",AT*(co.x-XCgrid->Middle.x)/fmtoGeVI,v[i].data[0]);
	  outUx.write(buffer,strlen(buffer));
	}
      if ((co.x==XCgrid->Middle.x)&&(co.z==XCgrid->Middle.z))
	{
	  sprintf(buffer, "%f\t %g\n",AT*(co.y-XCgrid->Middle.y)/fmtoGeVI,myresult);
	  outTy.write(buffer,strlen(buffer));
	  sprintf(buffer, "%f\t %g\n",AT*(co.y-XCgrid->Middle.y)/fmtoGeVI,v[i].data[1]);
	  outUy.write(buffer,strlen(buffer));
	}
       if ((co.x==XCgrid->Middle.x)&&(co.y==XCgrid->Middle.y))
	{
	  sprintf(buffer, "%f\t %g\n",AT*(co.z-XCgrid->Middle.z)/fmtoGeVI,myresult);
	  outTz.write(buffer,strlen(buffer));
	  sprintf(buffer, "%f\t %g\n",AT*(co.z-XCgrid->Middle.z)/fmtoGeVI,v[i].data[2]);
	  outUz.write(buffer,strlen(buffer));
	}
     
      if(co.z==XCgrid->Middle.z)
         {
           sprintf(buffer, "%f\t %g\n",AT*sqrt((co.x-XCgrid->Middle.x)*(co.x-XCgrid->Middle.x)+(co.y-XCgrid->Middle.y)*(co.y-XCgrid->Middle.y))/fmtoGeVI,sqrt(v[i].data[0]*v[i].data[0]+v[i].data[1]*v[i].data[1]));
           outvabs.write(buffer,strlen(buffer));
         }
       debug=0;
    }

 
  
  outTx.close();
  outUx.close();
  outTy.close();
  outUy.close();
  outTz.close();
  outUz.close();

  outTxy.close();
  outwarnxy.close();

  outvabs.close();
  
  

}

void BC(coord *co,xygrid *XCgrid)
{
   //periodic:
  
  while ((co->x)<0)
    co->x+=XCgrid->DIM.x;
  while(co->x>XCgrid->DIM.x-1)
    co->x-=XCgrid->DIM.x;
  
  while(co->y<0)
    co->y+=XCgrid->DIM.y;
  while(co->y>XCgrid->DIM.y-1)
    co->y-=XCgrid->DIM.y;
  
  while(co->z<0)
    co->z+=XCgrid->DIM.z;
  while(co->z>XCgrid->DIM.z-1)
    co->z-=XCgrid->DIM.z;
  
}

void loadeos()
{
  printf("===> Info: Loading/Calculating EoS...\t");
  const int length=10000;
  const double Tmin=0.01; //getting into numerical trouble for lower T
  const double Tmax=50.;
  double Tloc;
  double u0 [] = {0.,0.,0.,0.};

  double tab[(CDIM+1)*(CDIM+1)];

  double arrx[length];
  double arry[length];

  double arrst[length];

  xygrid feq;
  feq.init(SPEEDS,1,1);
  feq.alloc_data();
  
  for (int i=0;i<length;i++)
    {
      Tloc=Tmin+(Tmax-Tmin)/(length-1)*i;          
      setfeq(&feq,Tloc,u0,1.,0);      
      calcEMT(tab,&feq,0.);
      arrx[i]=Tloc;
      arry[i]=log(tab[0]);
      arrst[i]=tab[0]-tab[5];
      //printf("T=%f r=%f t00=%g\n",arrx[i],arry[i],tab[0]);
      //printf("T=%f r=%f t00=%g t11=%g\n",arrx[i],arrst[i],tab[0],tab[5]);
    }


  logeot.alloc(arry,arrx,length);
  enthalpy.alloc(arrx,arrst,length);
  
  for (int i=0;i<length;i++)
    {
      Tloc=Tmin+(Tmax-Tmin)/(length-1)*i;     
      setfeq(&feq,Tloc,u0,1.,0);
      calcEMT(tab,&feq,0.);
      arrx[i]=Tloc;
      arry[i]=log(tab[0]);
      arrst[i]=tab[0]-tab[5];
      //printf("test %g %g=?%g %g, T=%g=?%g\n",tab[0],enthalpy.f(Tofed(tab[0])),arrst[i],enthalpy.If(Tofed(tab[0])),Tofed(tab[0]),Tloc);
      //printf("test T=%g ed/T4=%g w/T4=%g Iw/T4=%g=%g\n",Tloc,tab[0]/pow(Tloc,4),enthalpy.f(Tloc)/pow(Tloc,4),enthalpy.If(Tloc)/pow(Tloc,5),(enthalpy.If(Tloc)+enthalpy.f(Tmin)*Tmin*0.2)/pow(Tloc,5));
    }

  //printf("test %f min %f\n",logeot.f(-80.),logeot.low);
  printf("finished\n");
}

double omegadyn(double temp)
{
  double res;
  double off=enthalpy.f(enthalpy.low)*enthalpy.low*0.5;
  res=(off+enthalpy.If(temp))/(ETAOS*enthalpy.f(temp)*temp);
  return res;
}

double omegadyn(double temp,gsl_interp_accel *acc)
{
  double res;
  double off=enthalpy.f(enthalpy.low,acc)*enthalpy.low*0.2;
  res=(enthalpy.If(temp,acc)+off)/(ETAOS*enthalpy.f(temp,acc)*temp);
  //printf("omegadyn temp=%f 1/res=%f en=%g Ien=%g\n",temp, 1/res,enthalpy.f(temp,acc),enthalpy.If(temp,acc)+off);

  return res;
}



void copydownf(xygrid *XCgrid)
{
  for (long int i=0;i<XCgrid->SIZE;i++)
    for (int j=0;j<SPEEDS;j++)
      {
	f[i].data[j]=fnew[i].data[j];
      }
}

//expects -1<=x<=1;
//returns -1,0,1
int mysign(double x)
{
  if (fabs(x)<1.e-15)
    return 0;
  else
    {
      if (x>0)
	return 1;
      else
	return -1;
    }
}

void printNvector(double *mat,int DIM,int flag)
{
  //standard
  if (flag==0)
    {
      printf("(");
      for (int b=0;b<DIM;b++)
	printf(" %g ",mat[b]);
      printf(")\n");
    }
  //Mathematica
  if (flag==1)
    {
      printf("{");
      for (int b=0;b<DIM-1;b++)
	printf("%g,",mat[b]);
      printf("%g}\n",mat[DIM-1]);
    }
}

void printNxNmatrix(double *mat,int DIM,int flag)
{
  //standard
  if (flag==0)
    {
      for (int a=0;a<DIM;a++)
	{
	  printf("(");
	  for (int b=0;b<DIM;b++)
	    printf(" %.16g ",mat[a*DIM+b]);
	  printf(")\n");
	}
    }

  //Mathematica
  if (flag==1)
    {
      printf("{");
      for (int a=0;a<DIM;a++)
	{
	  printf("{");
	  for (int b=0;b<DIM;b++)
	    {
	      printf("%f",mat[a*DIM+b]);
	      if (b<DIM-1)
		printf(",");
	    }
	  printf("}");
	  if(a<DIM-1)
	    printf(",");
	  else
	    printf("}");
	  printf("\n");
	}
    }
  

}

double interpolate_f(xygrid *XCgrid,xygrid *floc,coord * co,int j,int locald)
{
  coord coNN;

  double xx=co->x-dt*c[1][j]/c[0][j];
  double yy=co->y-dt*c[2][j]/c[0][j];
  double zz=co->z-dt*c[3][j]/c[0][j];
 
  int x0=int(floor(xx));
  int x1=int(ceil(xx));
  int y0=int(floor(yy));
  int y1=int(ceil(yy));
  int z0=int(floor(zz));
  int z1=int(ceil(zz));

  double dir[3];
  for (int a=0;a<3;a++)
    {
      if (mysign(c[a+1][j])==1)
	dir[a]=1-dt*c[a+1][j]/c[0][j];
      if (mysign(c[a+1][j])==-1)
	dir[a]=-dt*c[a+1][j]/c[0][j];
      if (mysign(c[a+1][j])==0)
	dir[a]=0;
    }
    

  
  long int help;

  coNN.set(x0,y0,z0);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f000=floc[help].data[j];

  coNN.set(x0,y0,z1);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f001=floc[help].data[j];

  coNN.set(x0,y1,z0);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f010=floc[help].data[j];

  coNN.set(x0,y1,z1);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f011=floc[help].data[j];
  
  coNN.set(x1,y0,z0);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f100=floc[help].data[j];

  
  coNN.set(x1,y0,z1);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f101=floc[help].data[j];

  coNN.set(x1,y1,z0);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f110=floc[help].data[j];

  coNN.set(x1,y1,z1);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f111=floc[help].data[j];

		   
  //printf("got till here for %f %f %f j=%d\n",xx,yy,zz,j);

  double temp;
  temp=trilinear_int(0,x1-x0, 0, y1-y0, 0, z1-z0,f000, f001,  f010,  f011, f100, f101,f110,f111,dir[0],dir[1],dir[2]);
  
  if (locald)
    {
      double mf=floc[XCgrid->getind(co)].data[j];
      printf("interpolate f at co.x=%i co.y=%i co.z=%i  j=%i f=%.16g\n",co->x,co->y,co->z,j,temp-mf);
      /*
      printf("j=%i f000=%.16g, f001=%.16g,  f010=%.16g,  f011=%.16g, f100=%.16g, f101=%.16g,f110=%.16g,f111=%.16g\n",j,f000,f001,f010,f011,f100,f101,f110,f111);
      printf("0 vals for %i %i %i j=%d\n",x0,y0,z0,j);
      //printf("floor values %i %i %i j=%d\n",int(floor(xx)),int(floor(yy)),int(floor(zz)),j);
      printf("1 vals for %i %i %i j=%d\n",x1,y1,z1,j);
      //printf("ceiling values %i %i %i j=%d\n",int(ceil(xx)),int(ceil(yy)),int(ceil(zz)),j);
      printf("evaluating at %f %f %f\n",xx,yy,zz);
      printf("final res %.16g vs %.16g where %.16g %.16g\n",temp,linear_int(0,1,f000,f100,dir[0]),xx,dir[0]);      */
      }



  return temp;
}

double finite_diff_f(xygrid *XCgrid,xygrid *floc,coord * co,int j,int locald,int difford)
{
  coord coNN;

 
 
  int x;
  long int help;

  double f000=floc[XCgrid->getind(co)].data[j];
  double dfdx,dfdy,dfdz;

  if (c[1][j]<0)
    coNN.set(co->x+1,co->y,co->z);
  else
    coNN.set(co->x-1,co->y,co->z);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f100=floc[help].data[j];

  dfdx=f000-f100;
  if (difford==2)
    {
      if (c[1][j]<0)
	coNN.set(co->x+2,co->y,co->z);
      else
	coNN.set(co->x-2,co->y,co->z);
      BC(&coNN,XCgrid);
      help=XCgrid->getind(&coNN);
      double f200=floc[help].data[j];
      dfdx=0.5*(3*f000-4*f100+f200);
    }

  if (c[1][j]<0)
    {
      dfdx*=-1;
    }

  if (c[2][j]<0)
    coNN.set(co->x,co->y+1,co->z);
  else
    coNN.set(co->x,co->y-1,co->z);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f010=floc[help].data[j];

  dfdy=f000-f010;
  if (difford==2)
    {
      if (c[2][j]<0)
	coNN.set(co->x,co->y+2,co->z);
      else
	coNN.set(co->x,co->y-2,co->z);
      BC(&coNN,XCgrid);
      help=XCgrid->getind(&coNN);
      double f020=floc[help].data[j];
      dfdy=0.5*(3*f000-4*f010+f020);
    }


  if (c[2][j]<0)
    {
      dfdy*=-1;
    }

  if (c[3][j]<0)
    coNN.set(co->x,co->y,co->z+1);
  else
    coNN.set(co->x,co->y,co->z-1);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  double f001=floc[help].data[j];

  dfdz=f000-f001;
  if (difford==2)
    {
      if (c[3][j]<0)
	coNN.set(co->x,co->y,co->z+2);
      else
	coNN.set(co->x,co->y,co->z+2);
      BC(&coNN,XCgrid);
      help=XCgrid->getind(&coNN);
      double f002=floc[help].data[j];
      dfdz=0.5*(3*f000-4*f001+f002);
    }


  if (c[3][j]<0)
    {
      dfdz*=-1;
    }


  double temp;
  temp=f000-dt*c[1][j]/c[0][j]*dfdx-dt*c[2][j]/c[0][j]*dfdy-dt*c[3][j]/c[0][j]*dfdz;

  if (locald)
    {
      printf("finite diff f at co.x=%i co.y=%i co.z=%i j=%i df=%.16g\n",co->x,co->y,co->z,j,temp-f000);
    }

  return temp;
}

double d2x_f(xygrid *XCgrid,xygrid *floc,coord * co,int j,int locald)
{
  coord coNN;
  int x;
  long int help;
  
  double f000=floc[XCgrid->getind(co)].data[j];
  double dfd2x,dfd2y,dfd2z;
  double fp,fm;
  
  coNN.set(co->x+1,co->y,co->z);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  fp=floc[help].data[j];
  coNN.set(co->x-1,co->y,co->z);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  fm=floc[help].data[j];

  dfd2x=fp-2*f000+fm;

  coNN.set(co->x,co->y+1,co->z);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  fp=floc[help].data[j];
  coNN.set(co->x,co->y-1,co->z);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  fm=floc[help].data[j];

  dfd2y=fp-2*f000+fm;

  coNN.set(co->x,co->y,co->z+1);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  fp=floc[help].data[j];
  coNN.set(co->x,co->y,co->z-1);
  BC(&coNN,XCgrid);
  help=XCgrid->getind(&coNN);
  fm=floc[help].data[j];

  dfd2z=fp-2*f000+fm;

  return (dfd2z+dfd2y+dfd2x);
}


double fluxlim_f(xygrid *XCgrid,xygrid *floc,coord * co,int j,int locald)
{
  coord coNN;
  int x;
  long int help;

  double f000=floc[XCgrid->getind(co)].data[j];

  double dfdx,dfdy,dfdz;
  double fluxplus,fluxminus,sigma;
  double known=0*dt;

  if (fabs(c[1][j])>1.e-6)
    {
      if (c[1][j]<0)
	{
	  coNN.set(co->x+1,co->y,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fp=floc[help].data[j];
	  coNN.set(co->x-1,co->y,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fm=floc[help].data[j];
	  coNN.set(co->x+2,co->y,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fpp=floc[help].data[j];
	  sigma=minmod(f000-fm,fp-f000);
	  fluxminus=(f000-0.5*sigma*(1+known*c[1][j]/c[0][j]));      
	  sigma=minmod(fp-f000,fpp-fp);
	  fluxplus=(fp-0.5*sigma*(1+known*c[1][j]/c[0][j]));
	}
      else
	{
	  coNN.set(co->x+1,co->y,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fp=floc[help].data[j];
	  coNN.set(co->x-1,co->y,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fm=floc[help].data[j];
	  coNN.set(co->x-2,co->y,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fmm=floc[help].data[j];
	  sigma=minmod(fm-fmm,f000-fm);
	  fluxminus=(fm+0.5*sigma*(1-known*c[1][j]/c[0][j]));     
	  sigma=minmod(f000-fm,fp-f000);
	  fluxplus=(f000+0.5*sigma*(1-known*c[1][j]/c[0][j]));	
	}
    }
  else
    {
      fluxplus=0;
      fluxminus=0;
    }
  
  
  dfdx=fluxplus-fluxminus;
  
  if (fabs(c[2][j])>1.e-6)
    {
      if (c[2][j]<0)
	{
	  coNN.set(co->x,co->y+1,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fp=floc[help].data[j];
	  coNN.set(co->x,co->y-1,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fm=floc[help].data[j];
	  coNN.set(co->x,co->y+2,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fpp=floc[help].data[j];
	  sigma=minmod(f000-fm,fp-f000);
	  fluxminus=(f000-0.5*sigma*(1+known*c[2][j]/c[0][j]));      
	  sigma=minmod(fp-f000,fpp-fp);
	  fluxplus=(fp-0.5*sigma*(1+known*c[2][j]/c[0][j]));
	}
      else
	{
	  coNN.set(co->x,co->y+1,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fp=floc[help].data[j];
	  coNN.set(co->x,co->y-1,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fm=floc[help].data[j];
	  coNN.set(co->x,co->y-2,co->z);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fmm=floc[help].data[j];
	  sigma=minmod(fm-fmm,f000-fm);
	  fluxminus=(fm+0.5*sigma*(1-known*c[2][j]/c[0][j]));     
	  sigma=minmod(f000-fm,fp-f000);
	  fluxplus=(f000+0.5*sigma*(1-known*c[2][j]/c[0][j]));	
	}
    }
  else
    {
      fluxplus=0;
      fluxminus=0;
    }
  
  dfdy=fluxplus-fluxminus;

  if (fabs(c[3][j])>1.e-6)
    {
      if (c[3][j]<0)
	{
	  coNN.set(co->x,co->y,co->z+1);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fp=floc[help].data[j];
	  coNN.set(co->x,co->y,co->z-1);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fm=floc[help].data[j];
	  coNN.set(co->x,co->y,co->z+2);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fpp=floc[help].data[j];
	  sigma=minmod(f000-fm,fp-f000);
	  fluxminus=(f000-0.5*sigma*(1+known*c[3][j]/c[0][j]));      
	  sigma=minmod(fp-f000,fpp-fp);
	  fluxplus=(fp-0.5*sigma*(1+known*c[3][j]/c[0][j]));
	}
      else
	{
	  coNN.set(co->x,co->y,co->z+1);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fp=floc[help].data[j];
	  coNN.set(co->x,co->y,co->z-1);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fm=floc[help].data[j];
	  coNN.set(co->x,co->y,co->z-2);
	  BC(&coNN,XCgrid);
	  help=XCgrid->getind(&coNN);
	  double fmm=floc[help].data[j];
	  sigma=minmod(fm-fmm,f000-fm);
	  fluxminus=(fm+0.5*sigma*(1-known*c[3][j]/c[0][j]));     
	  sigma=minmod(f000-fm,fp-f000);
	  fluxplus=(f000+0.5*sigma*(1-known*c[3][j]/c[0][j]));	
	}
    }
  else
    {
      fluxplus=0;
      fluxminus=0;
    }
  
  dfdz=fluxplus-fluxminus;


  double temp;
  temp=f000-dt*c[1][j]/c[0][j]*dfdx-dt*c[2][j]/c[0][j]*dfdy-dt*c[3][j]/c[0][j]*dfdz;

  if (locald)
    {
      printf("finite diff f at co.x=%i co.y=%i co.z=%i j=%i df=%.16g\n",co->x,co->y,co->z,j,temp-f000);
    }

  return temp;
}


void dotheludwig(xygrid *XCgrid)
{
  
  //double xval,dist;
  coord co,coNN,cohelp;
  long int ihigh;
  

  double stat[4];
  double holdmat[16];
  double coeffs[numNpoly];

  
  xygrid feq,fhold;
  
  double help,u0temp,Ttemp;
  double xval;

  long int i;
  int j;
  int a;
  int nthreads, tid;
  gsl_interp_accel *acc;
  int WARN;

  omp_set_num_threads(num_threads);

  int perform[num_threads];
  int maxval;


  gsl_vector_complex *eval;
  gsl_matrix_complex *evec;
  gsl_matrix_view m;
  gsl_eigen_nonsymmv_workspace *WS;
  gsl_vector_view revec;
  gsl_vector_view reval;

#pragma omp parallel shared(f,fnew,Tab,T,v,u0,warn,nthreads,perform,dist,ABORT) private(i,j,a,tid,help,u0temp,Ttemp,stat,feq,fhold,co,coNN,cohelp,ihigh,acc,eval,evec,m,WS,revec,reval,holdmat,xval,maxval,coeffs,debug)
  {
    debug=0;
    acc=gsl_interp_accel_alloc (); 

    feq.init(SPEEDS,1,1);
    feq.alloc_data();
    
    fhold.init(SPEEDS,1,1);
    fhold.alloc_data();
    

    eval = gsl_vector_complex_alloc (4);
    evec = gsl_matrix_complex_alloc (4, 4);
    WS = gsl_eigen_nonsymmv_alloc (4);
    
    tid = omp_get_thread_num();
    if (tid == 0)
      {
	nthreads = omp_get_num_threads();
	//printf("===> Info: Number of threads = %d\n", nthreads);
      }   
    //perform[tid]=0;
    //printf("===> Info: Thread %d starting... performed=%i\n",tid,perform[tid]);

#pragma omp for schedule(dynamic,1)
    for (i=0;i<XCgrid->SIZE;i++)
      {
	warn[i]=0;
	XCgrid->getcoord(&co,i);
	//printf("doing i=%li\n",i);

	/*
	if ((co.x==XCgrid->Middle.x)&&(co.y==XCgrid->Middle.y+104))
	  debug=1;
	if ((co.x==XCgrid->Middle.x+104)&&(co.y==XCgrid->Middle.y))
	  debug=1;
	*/

	if (debug)
	  printf("\t debugging at x=%f y=%f(co.x=%i co.y=%i)\n",AT*(co.x-XCgrid->Middle.x)/fmtoGeVI,AT*(co.y-XCgrid->Middle.y)/fmtoGeVI,co.x,co.y);
	
	/*
	if (((co.x==XCgrid->Middle.x+78)||(co.x==XCgrid->Middle.x-78))&&(co.y==XCgrid->Middle.y))
	  {
	    printf("\t debugging at %f (co.x=%i co.y=%i)\n",AT*(co.x-XCgrid->Middle.x)/fmtoGeVI,co.x,co.y);
	    debug=1;
	    }*/

	/*
	if (((co.x==XCgrid->Middle.x-79)||(co.x==XCgrid->Middle.x+79))&&(co.y==XCgrid->Middle.y))
	  {
	    printf("\t debugging at %f (co.x=%i co.y=%i)\n",AT*(co.x-XCgrid->Middle.x)/fmtoGeVI,co.x,co.y);
	    debug=1;
	    }*/

	//setfeq(&fhold,T[i].data[0],v[i].data,u0[i].data[0],debug); //stability purposes, may be removed

	calcEMT(Tab[i].data,&f[i],EDreg);

	
	if (debug)
	  {
	    //printNxNmatrix(Tab[i].data,4,0);
	  }

	//warn[i]=calcedvnew(Tab[i].data,stat);
	warn[i]=calcedvquick(Tab[i].data,stat,0);


	if (!warn[i])
	  {	    	    
	    for (a=0;a<CDIM;a++)
	      {
		v[i].data[a]=stat[a+1];						
	      }

	    T[i].data[0]=Tofed(stat[0],acc);
	    u0[i].data[0]=sqrt(1./(1-v[i].data[0]*v[i].data[0]-v[i].data[1]*v[i].data[1]-v[i].data[2]*v[i].data[2]));

	  }
	else
	  {
	    if (OUTPUTWARN)
	      {
		printf("warn=%i at i=%li x=%f\n",warn[i],i,AT*(co.x-XCgrid->Middle.x)/fmtoGeVI);
		printNxNmatrix(Tab[i].data,4,1);
	      }
	    //leave hydro variables untouched
	  }

	/*
	if ((co.y==XCgrid->Middle.y)&&(fabs(v[i].data[1])>1.e-6))
	  printf("see there %i\n",co.x-XCgrid->Middle.x);
	*/

	
	if (debug)
	  {
	    printf("\tWARN=%i\n",warn[i]);
	    printf("\t(vx,vy,vz)=(%g,%g,%g)\n",v[i].data[0],v[i].data[1],v[i].data[2]);
	    printf("\tT=%g\n",T[i].data[0]);
	    }

	if ((isnan(u0[i].data[0]))||(isinf(u0[i].data[0])))
	  {
	    ABORT=1;
	    if (!NANWARN)
	      printf("ERROR @ i=%li in u0 v1=%g v2=%g v3=%g\n",i,v[i].data[0],v[i].data[1],v[i].data[2]);
	    warn[i]=1;
	    NANWARN=1;
	  }	
	if ((OUTPUTWARN)&&(warn[i]))
	  printf("i=%li T=%f\n",i,T[i].data[0]);

	
	setfeq(&feq,T[i].data[0],v[i].data,u0[i].data[0],debug);
	
	for (j=0;j<SPEEDS;j++)       
	  {
	    
	    //help=u0[i].data[0]*omegadyn(T[i].data[0],acc)*T[i].data[0]*AT*dt*(1-v[i].data[0]*c[1][j]/c[0][j]-v[i].data[1]*c[2][j]/c[0][j]-v[i].data[2]*c[3][j]/c[0][j]);

	    help=u0[i].data[0]/(5*ETAOS)*T[i].data[0]*AT*dt*(1-v[i].data[0]*c[1][j]/c[0][j]-v[i].data[1]*c[2][j]/c[0][j]-v[i].data[2]*c[3][j]/c[0][j]);
	    //help=0;



	    

	    
	    //this only works if sg33(t)=1/tau<1, otherwise not clear that interpolation to NN is appropriate
	   
	    //fnew[i].data[j]=(1-help)*f[i].data[j]+help*(2*feq.data[j]-fhold.data[j])+ bulkreg*dt*d2x_f(XCgrid,f,&co,j,0); //does not improve stability
	    
	    //fnew[i].data[j]=(1-help)*f[i].data[j]+help*(feq.data[j])+ bulkreg*dt*d2x_f(XCgrid,f,&co,j,0); 

	    fnew[i].data[j]=(1-help)*f[i].data[j]+help*(feq.data[j]);
	    if (ORDER==2) //only calculate half-step
	      {
		fnew[i].data[j]=(1-0.5*help)*f[i].data[j]+0.5*help*feq.data[j];
#ifdef INTERPOLATE
		fnew[i].data[j]+=0.5*(interpolate_f(XCgrid,f,&co,j,debug)-f[i].data[j]);
#endif
#ifdef UPWIND
		fnew[i].data[j]+=0.5*(finite_diff_f(XCgrid,f,&co,j,debug,LORDER)-f[i].data[j]);
#endif
#ifdef MINMOD
		fnew[i].data[j]+=0.5*(fluxlim_f(XCgrid,f,&co,j,debug)-f[i].data[j]);
#endif
	      }


	    if ((fnew[i].data[j]<0)&&(!NEGFWARN)&&(j!=SPEEDS-1))
	      {		
		printf("negative fnew @ i=%li fnew=%f f=%f help=%f feq=%f\n",i,fnew[i].data[j],f[i].data[j],help,feq.data[j]);
		printf("T=%f v=%f,%f u0=%f\n",T[i].data[0],v[i].data[0],v[i].data[1],u0[i].data[0]);
		NEGFWARN=1;
	      }

	    if (debug)
	      {
		//if (j==1||j==7)
		{
		  printf("\t f[%i]=%.16g, fnew=%.16g\n",j,f[i].data[j],fnew[i].data[j]);
		  printf("help=%g, feqdata=%g\n",help,feq.data[j]);
		}
	      }

	  

	    
	    	    
	    if (isnan(fnew[i].data[j]))
	      {
		if (!NANWARN)
		  printf("ERROR isnan fnew @ i=%li\n",i);
		NANWARN=1;
		warn[i]=1;
	      }
	  }


#ifdef MILNE
	setdfdz1(&feq,&f[i],0);	
	for (j=0;j<SPEEDS;j++)
	  fnew[i].data[j]+=AT*dt*sg33(t)*feq.data[j];	  
#endif
       	
	debug=0;
	
	
      }
    
    


#pragma omp for schedule(dynamic,1)
    for (long int i=0;i<XCgrid->SIZE;i++)
      {
	XCgrid->getcoord(&co,i);
	if (ORDER==1)
	  {	    
	    for (int j=0;j<SPEEDS;j++)
	      {
#ifdef INTERPOLATE
		f[i].data[j]=interpolate_f(XCgrid,fnew,&co,j,debug);
#endif
#ifdef UPWIND
		f[i].data[j]=finite_diff_f(XCgrid,fnew,&co,j,debug,LORDER);
#endif
#ifdef MINMOD
		f[i].data[j]=fluxlim_f(XCgrid,fnew,&co,j,debug);
#endif
	      }
	  }

	if (ORDER==2)
	  {
	    calcEMT(Tab[i].data,&fnew[i],EDreg);
	    //warn[i]=calcedvnew(Tab[i].data,stat);
	    warn[i]=calcedvquick(Tab[i].data,stat,0);
	    Ttemp=Tofed(stat[0],acc);
	    u0temp=sqrt(1./(1-stat[1]*stat[1]-stat[2]*stat[2]-stat[3]*stat[3]));
	    setfeq(&feq,Ttemp,&stat[1],u0temp,0);
	    for (int j=0;j<SPEEDS;j++)
	      {
#ifdef INTERPOLATE
		f[i].data[j]+=(interpolate_f(XCgrid,fnew,&co,j,debug)-fnew[i].data[j]);
#endif
#ifdef UPWIND
		f[i].data[j]+=(finite_diff_f(XCgrid,fnew,&co,j,debug,LORDER)-fnew[i].data[j]);
#endif
#ifdef MINMOD
		f[i].data[j]+=(fluxlim_f(XCgrid,fnew,&co,j,debug)-fnew[i].data[j]);
#endif
		help=u0temp/(5*ETAOS)*Ttemp*AT*dt*(1-stat[1]*c[1][j]/c[0][j]-stat[2]*c[2][j]/c[0][j]-stat[3]*c[3][j]/c[0][j]);
		f[i].data[j]+=(-help)*(fnew[i].data[j]-feq.data[j]);
		if (isnan(f[i].data[j]))
		  printf("ERROR isnan fn ORDER 2 @ i=%li\n",i);
	      }
	  }
	  
      }


    gsl_eigen_nonsymmv_free (WS);
    gsl_vector_complex_free (eval);
    gsl_matrix_complex_free (evec);
    
    gsl_interp_accel_free (acc);
  }

  

//end of omp routine
  // for (int l=0;l<num_threads;l++)
  // if (perform[l]<10000)
  //   printf("underperform @%i (%i)\n",l,perform[l]);
  
  //copydownf(XCgrid);
}




int main(int argc, char* argv[])
{

#ifdef D3Q1_6_3
  sprintf(desc,"D3S18");
#endif 
#ifdef D3Q1_6_3_plus
  sprintf(desc,"D3S19");
#endif 
#ifdef D3Q1_8_3
  sprintf(desc,"D3S25");
#endif 
#ifdef D3Q1_8_3_plus
  sprintf(desc,"D3S26");
#endif
#ifdef D3Q1_8_4_plus
  sprintf(desc,"D3S33");
#endif
#ifdef D3Q1_8_5_plus
  sprintf(desc,"D3S41");
#endif
#ifdef D3Q18
  sprintf(desc,"D3Q18");
#endif 
#ifdef D3Q26
  sprintf(desc,"D3Q26");
#endif 
  
  
  char *paramsfile;
  
  if (argc>1)
    {
      paramsfile=argv[1];
    }
  else
    {
      printf("ERROR please specify a params file name");
    }

  fstream parameters;
  parameters.open(paramsfile, ios::in);
  if(parameters.is_open())
    {
      int dummyint;
      char dummychar[255];
      double dummydouble;     
      parameters >> dummychar;
      parameters >> dummyint;
      //printf("Found NX=%i\n",dummyint);
      NX=dummyint;
      parameters >> dummychar;
      parameters >> dummyint;
      NY=dummyint;
      //printf("Found NY=%i\n",dummyint);
      parameters >> dummychar;
      parameters >> dummyint;
      NZ=dummyint;      
      //printf("Found =%f\n",dummydouble);
      parameters >> dummychar;
      parameters >> dummydouble;
      AT=dummydouble;
      
      parameters >> dummychar;
      parameters >> dummydouble;
      dt=dummydouble;
      //printf("Found =%f\n",dummydouble);
      parameters >> dummychar;
      parameters >> dummydouble;
      ETAOS=dummydouble;
      //printf("Found NZ=%i\n",dummyint);
      parameters >> dummychar;
      parameters >> dummydouble;
      TINIT=dummydouble;
      parameters >> dummychar;
      parameters >> dummydouble;
      alpha=dummydouble;
      parameters >> dummychar;
      parameters >> dummydouble;
      regulator=dummydouble;
      parameters >> dummychar;
      parameters >> dummydouble;
      regsigma=dummydouble;
      parameters >> dummychar;
      parameters >> dummyint;
      SNAPOUT=dummyint;
      parameters >> dummychar;
      parameters >> dummyint;
      ORDER=dummyint;
      parameters >> dummychar;
      parameters >> dummydouble;
      dof=dummydouble;
      //printf("Found dof=%f\n",dof);
    }
  else
    {
      printf("\nERROR: params file %s could not be opened\n",paramsfile);
      ABORT=1;
    }
  parameters.close();
  //printf("time discretization order = %i\n",ORDER);

  //still need to set outdir
  //sprintf(cbuffer,"data");
  //sprintf(cbuffer,"%i-%i-%i-A%.3f-dt%.4f-ETA%.4f-reg%.2f-sig%.3f-alpha%.1f-bulk%.3f-10a%.1f-%s-O%i",NX,NY,NZ,AT,dt,ETAOS,regulator,regsigma,alpha,bulkreg,a10,desc,ORDER);
  sprintf(cbuffer,"%i-%i-%i-A%.3f-tini%.2f-dof%.1f-dt%.4f-ETA%.4f-%s-O%i",NX,NY,NZ,AT,TINIT,dof,dt,ETAOS,desc,ORDER);
  outdir=cbuffer;
  
  //test if directory exists 

  struct stat st;
  
  if (stat(outdir,&st) != 0)
    {
      printf("===> Info: directory %s not present, creating it\n",outdir);
      mkdir(outdir,S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
    }
 
  char tmpchar[1255];

  sprintf(tmpchar,"cp %s %s",paramsfile,outdir);
  //printf("have %s\n",tmpchar);
  system(tmpchar);

  num_threads=omp_get_max_threads();
  printf("===> Info: maximum number of threads=%i,",num_threads);
  //num_threads=1;
  printf(" using %i\n",num_threads);
  
  xygrid space(NX,NY,NZ);
  
  allocate_memory(&space);
  
  setspeeds();
  
  t=1./AT; //to get eos running

  loadeos();

  initialize(&space);

  printf("===> Info: alpha=%f\n",alpha);
  printf("===> Info: dt=%f\n",dt);
  printf("===> Info: ETA/S=%f\n",ETAOS);
  printf("===> Info: regulator=%f\n",regulator);
  printf("===> Info: sigma=%f\n",regsigma);
  
  //printf("testing: %f %f\n",omegadyn(0.5),omegadyn(0.2));
 
  /*
  //test:
  double Tab[16];
  xygrid feq,dfdz;
  feq.init(SPEEDS,1,1);
  feq.alloc_data();
  dfdz.init(SPEEDS,1,1);
  dfdz.alloc_data();
  double v[]={0.49,0.0,0.0};

  debug=1;
  setfeq(&feq,2.,v,sqrt(1/(1-v[0]*v[0]-v[1]*v[1]-v[2]*v[2])),debug);
  debug=0;
  for (int i=0;i<SPEEDS;i++)
    printf("i=%i f=%f\n",i,feq.data[i]);
  */
  /*
  setdfdz1(&dfdz,&feq,1);
  */

  /*
  for (int i=0;i<SPEEDS;i++)
    {     
      double expweight=exp(-1.0*c[0][i]);
      //feq.data[i]=expweight*poly[5][i]*poly[0][i];
      //feq.data[i]=expweight*c[0][i]*c[1][i]*c[1][i];
      }*/

  /*
  double test=0;
  for (int i=0;i<SPEEDS;i++)
    test+=w[i]*dfdz.data[i];

  printf("TEST %g\n",test);
  */
 
  
  //printf("feqsimple %i\n",feqsimple);

  /*
  calcEMT(Tab,&feq,1.e-7);
  printf("TEST Tab:\n");
  printNxNmatrix(Tab,4,1);
  
  double stat[4];
  
  debug=1;
  calcedvnew(Tab,stat);
  debug=0;
 
  printf("aha T=%g v1=%g v2=%g v3=%g\n",Tofed(stat[0]),stat[1],stat[2],stat[3]);
  */
  //finished test


  //double Tab[16];
  //calcEMT(Tab,&f[space.intMiddle]);
  //printf("T00 center =%g corresponding T=%g\n",Tab[0],Tofed(Tab[0]));
  /*
  coord co;
  double Tab[3];
  for (long int i=0;i<space.SIZE;i++)
    {
      space.getcoord(&co,i);
      calcEMT(Tab,&f[i]);
      printf("i=%li T00=%g T=%f\n",i,Tab[0],Tofed(Tab[0]));
      }*/

  //tests();
 
  //outputT0(&space);
  /*
  printf("test 5->%i\n",BC(5,&space));
  printf("test -5->%i\n",BC(-5,&space));
  printf("test 0->%i\n",BC(0,&space));
  printf("test -1->%i\n",BC(-1,&space));
  printf("test 111->%i\n",BC(111,&space));
  */
  
  sprintf(tmpchar,"%s/meta.dat",outdir);
  meta.open(tmpchar, ios::out);
  sprintf(tmpchar,"#Time [fm]\t ED[fm^-4]\t Pi [fm^-4]  Sigma [fm^-1\n");
  meta.write(tmpchar,strlen(tmpchar));


  printf("===> Info: Starting main routine...\n");

  /*
  double Tab[16];
  calcEMT(Tab,&f[space.intMiddle]);
  printf("tab0 =%g\n",Tab[0]);
    double sum=0;
  for (int i=0;i<SPEEDS;i++)
    sum+=f[space.intMiddle].data[i]*w[i]*c[0][i]*c[0][i];
    printf("checking feq %g\n",sum);*/


  t=TINIT*fmtoGeVI/AT;
  
  long int i=0;

  double tstop=4.1;
  printf("Stopping at %f fm/c",tstop);
  while((t*AT<tstop*fmtoGeVI)&&(!ABORT))
    {

      dotheludwig(&space);
      if ((i%SNAPOUT==0)||(ABORT))
	output(&space);

      i++; 
      t+=dt;
      if (ABORT)
	printf("===> Info: got ABORT signal, trying to exit gracefully \n");

    }
  
  free_memory();

  meta.close();
   if (!ABORT)
     printf("Finished successfully\n");
   else
     printf("Finished with errors\n");
}
