#include <math.h>
struct state_vec
{
  double umu[4];
  double ed;
  double T;
  double M4;
  double Tab[4][4]; //EMT components
  int isfluid;  //nonzero if fluid 
  
};

struct coord
{
  int x,y,z;

  void print()
  {
    printf("x=%i y=%i z=%i\n",x,y,z);
  }
  
  void set(int a,int b, int c)
  {
    x=a;
    y=b;
    z=c;
  }
  
  double veclen()
  {
    return sqrt(x*x+y*y+z*z);
  }

};




class data_array
{
 public:
  
   struct coord DIM;

   long int SIZE;

   //generic data pointer
   //should contain double array of size SIZE 
   double * data;

   bool dat_alloc;

   data_array()
     {
       dat_alloc=0;
     };

   ~data_array()
     {
       //printf("===> Info :  Deleting data_array\n");
       //if we allocated data, free memory
       if (dat_alloc)
	 delete [] data;
     };

   void init(int Nx,int Ny,int Nz);
   void init(struct coord tdim);

   void alloc_data();
};




class xygrid: public data_array
{

 
  
 public:

  xygrid()
    {
      dat_alloc=0;
    };

  xygrid(int Nx,int Ny,int Nz);

  //struct coord DIM;

  struct coord Middle;

  long int intMiddle;
  //this will hold total number of points
  //long int SIZE;


  double dt,a;

  //generic data pointer
  //should contain double array of size SIZE 
  //double * data;


  long int getind(int sx,int sy, int sz);
  long int getind(struct coord *co);
  void getcoord(struct coord * co,long int ind);
};


xygrid::xygrid(int Nx,int Ny,int Nz)
{

  DIM.x=Nx;
  DIM.y=Ny;
  DIM.z=Nz;

  printf("===> Info: Generating %ix%ix%i grid\n",Nx,Ny,Nz);

  SIZE=Nx*Ny*Nz;

  Middle.x=((Nx-1)/2);
  Middle.y=((Ny-1)/2);
  Middle.z=((Nz-1)/2);
  
  intMiddle=getind(&Middle);

  dat_alloc=0;
}


long int xygrid::getind(int sx,int sy, int sz)
{
  long int temp;
  temp=sx*DIM.y*DIM.z+sy*DIM.z+sz;

  if ((temp<0)||(temp>=SIZE))
    printf("WARNING (structs:getind) : index out of range\n");

  return temp;
}


long int xygrid::getind(struct coord * co)
{
  long int temp;
  temp=(co->x)*DIM.y*DIM.z+(co->y)*DIM.z+(co->z);

  if ((temp<0)||(temp>=SIZE))
    printf("WARNING (structs:getind) : index out of range\n");
  return temp;
}


void xygrid::getcoord(struct coord * co,long int ind)
{
  long int temp;
  (*co).z=ind%(DIM.z);
  temp=((ind-(*co).z)/DIM.z);
  (*co).y=temp%(DIM.y);
  (*co).x=temp/(DIM.y);
}

void data_array::init(int Nx,int Ny,int Nz)
{

  DIM.x=Nx;
  DIM.y=Ny;
  DIM.z=Nz;

  //printf("===> Info: Generating %ix%ix%i data_array\n",Nx,Ny,Nz);

  SIZE=Nx*Ny*Nz;

  dat_alloc=0;
}

void data_array::init(struct coord tdim)
{
  DIM.x=tdim.x;
  DIM.y=tdim.y;
  DIM.z=tdim.z;

  //printf("===> Info: Generating %ix%ix%i data_array\n",DIM.x,DIM.y,DIM.z);

  SIZE=DIM.x*DIM.y*DIM.z;

  dat_alloc=0;
}


void data_array::alloc_data()
{
  data = new double[SIZE];
  dat_alloc=1;
}

